import os
from pathlib import Path
import typing as t

from bestconfig.source import Source
from bestconfig.file_parsers import *


class AbstractAdapter(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def get_dict(cls, source: Source) -> dict:
        """Возвращает словарь с конфигами,
        необходимые параметры находятся в source"""
        pass


class EnvAdapter(AbstractAdapter):
    """Адаптер для доступа к переменным окружения"""

    @classmethod
    def get_dict(cls, source: Source) -> dict:
        return dict(os.environ)


class DictAdapter(AbstractAdapter):
    """Получение конфига из python dict"""

    @classmethod
    def get_dict(cls, source: Source) -> dict:
        return source.get('data')


class FileAdapter(AbstractAdapter):
    """Читает и пишет в файл"""

    @classmethod
    def get_dict(cls, source: Source) -> dict:
        filepath = source.get('filepath')
        assert isinstance(filepath, Path)
        filetype = cls._get_file_type(filepath)

        if filetype in cls.specific_parsers:
            parser = cls.specific_parsers[filetype]
        else:
            raise NotImplementedError('This file type does not supported yet %s' % filepath)
        return parser.read(str(filepath))

    # TODO добавить .env .cfg .ini
    """Обработчики файлов нужного типа"""
    specific_parsers = {
        'json': JsonParser,
        'yaml': YamlParser,
        'ini': IniParser,
        # same with .ini
        'cfg': IniParser,
        'env': EnvParser,
        'py': PyParser,
    }

    """Общепринятые названия и соответствующие им расширения"""
    file_types = {
        'env_file': 'env',
        '.env': 'env',
    }

    @classmethod
    def _get_file_type(cls, filepath: Path) -> t.Optional[str]:
        """Возвращает тип файла, обычно просто его расширение
        None если не найдено"""
        _, ext = os.path.splitext(filepath)
        filename = os.path.basename(filepath)

        ext = ext.strip('.')
        if ext in cls.specific_parsers:
            return ext

        if filename in cls.file_types:
            return cls.file_types[filename]

        return None



from warnings import warn
import inspect
from bestconfig.converters import *
from bestconfig.source_resolver import SourceResolver, FilesScanner
from bestconfig.source import TargetType

"""Тип, которым может быть значение конфига"""
ConfigType = t.Union[str, int, float, bool, dict, list]

"""
Используется по умолчанию, при
config.get()
Может быть переопределен параметром 'cast'
"""
default_converter = UniversalConverter()


class ConfigProvider(dict):
    """
    Основной класс, с которым имеет дело пользователь
    после создания конфига.
    Предоставляет универсальный интерфейс для хранящихся данных
    Пример:
    config = Config()
    assert isinstance(config, ConfigProvider)
    print(config.logger.format)
    print(config['logger.format'])
    print(config['logger']['format'])
    print(config.logger['format'])
    print(config.get('logger'))
    """

    def __init__(self, data: dict):
        super().__init__(data)

    def get(self, item: str, default_value=None, raise_absent=False,
            cast: t.Optional[AbstractConverter] = default_converter) -> ConfigType:
        """
        Обращается к self._data и возвращает собственный экземпляр, если значение dict
        в противном случае само значение

        attr_name = config.get('attr_name', 123)

        :param item: ключ, значение по которому нужно вернуть
        :param default_value: при отсутствии ключа, возвращается оно
        :param raise_absent: если ключа не существует, кинуть KeyError,
        в противном случает вернуть None
        :param cast: наследник класса AbstractConverter,
        если передан, возвращается cast.cast(value)
        :return: значение по ключу, None или KeyError
        """
        assert isinstance(item, str), 'Key must be str, not %s' % type(item)

        try:
            # Обработка случая config.get('key.other')
            value = self._unsafe_access_key(item)
            # Возвращаем словарь в виде класса ConfigProvider
            if isinstance(value, dict):
                return self.__class__(value)

            # Преобразуем объект в соответствии с переданным в параметрах cast
            if cast:
                return cast.cast(value)
            return value
        except KeyError:
            if raise_absent:
                raise KeyError

        return default_value

    def get_raw(self, item: str):
        """Возвращает значение по ключу, не изменяя его
        Такое, какое было считано из файла"""
        return self.get(item, cast=None)

    def contains(self, item: str):
        return self.get(item) is not None

    def assert_contains(self, item: str):
        """Бросает исключение KeyError, если ключ не найден"""
        self.get(item, raise_absent=True)

    def set(self, item: str, value: ConfigType):
        """Устанавливает значение по ключу"""
        if not item:
            warn('Использование пустой строки в качестве ключа', UserWarning)
        self[item] = value

    def to_dict(self) -> dict:
        """Возвращает весь конфигурационные словарь, содержащий имеющиеся данные
        Эквивалентно dict(config)"""
        return dict(self)

    def insert(self, target: TargetType):
        """
        Добавляет в основное хранилище новые данные
        по указанию target (то же самое, что аргумент в Config())
        :param target: имя файла или словарь
        """
        resolver = SourceResolver(FilesScanner.get_caller_path())
        new_data = resolver.resolve(target)
        self.update(new_data)

    def update_from_locals(self):
        """Обновляет словарь отфильтрованными локальными переменными
        из вызвавшего функцию контекста.
        Переменные только типов (str, int, float, dict, list)
        будут учтены
        """
        config_types = (str, int, float, dict, list)

        def is_config_var(name, value):
            if not isinstance(value, config_types):
                return False
            if name.startswith('__'):
                return False
            if isinstance(value, self.__class__):
                return False
            return True

        local_vars = inspect.currentframe().f_back.f_locals
        data = {
            key: value for key, value in local_vars.items()
            if is_config_var(key, value)
        }
        self.insert(data)

    def int(self, item: str) -> t.Optional[int]:
        """ config.int('limit') -> 45 """
        return self.get(item, cast=SimpleConverter(int))

    def float(self, item: str) -> t.Optional[float]:
        """ type(config.float('wait_time')) in [float, None] """
        return self.get(item, cast=SimpleConverter(float))

    def bool(self, item: str) -> t.Optional[float]:
        """ config.bool('DEBUG') in [True, False, None] """
        # Тут используется не универсальный преобразователь
        return self.get(item, cast=BoolConverter())

    def list(self, item: str) -> t.Optional[list]:
        """ config.list('items') -> [1, 3, 4] """
        return self.get(item, cast=PythonicConverter(empty_as=[]))

    def dict(self, item: str) -> t.Optional[dict]:
        """ config.dict('logger') -> {'mode': 'WARNING'} """
        return self.get(item, cast=PythonicConverter(empty_as=dict()))

    def str(self, item: str) -> t.Optional[dict]:
        return self.get(item, cast=SimpleConverter(str))

    def __getattr__(self, item):
        """attr_name = config.attr_name"""
        return self.get(item, raise_absent=True)

    def __getitem__(self, item):
        """attr_name = config['attr_name']"""
        return self.get(item, raise_absent=True)

    def __contains__(self, item: str):
        return self.contains(item)

    def _unsafe_access_key(self, item: str) -> t.Optional[ConfigType]:
        """Возвращает значение из _data, пытаясь его найти
        по строке виде key.subkey.otherkey или без точки
        Пример: config.get('logger.mode')
        config.get('ADMIN_ID')
        Кидает KeyError, при отсутствии ключа
        """

        if super().__contains__(item):
            return super().__getitem__(item)

        # Попытка пройти по частям ключей, разделенным точками
        keys = item.split('.')
        value = self.to_dict()
        for key in keys:
            if not hasattr(value, '__getitem__'):
                raise KeyError

            if key not in value:
                raise KeyError

            value = value[key]

        return value



import typing as t
from bestconfig.config_provider import ConfigProvider
from bestconfig.source import Source, TargetType
from bestconfig.source_resolver import SourceResolver, FilesScanner

supported_extensions = ['json', 'yaml', 'ini', 'cfg', 'env']
applicant_files = ['config', 'conf', 'setting', 'settings', 'configuration']


def generate_targets(files: t.List[str], extensions: t.List[str]) -> t.List[str]:
    """Генерирует всевозможные названия файлов
    на основе предоставленных претендентов на названия и расширения"""
    return list(sum(
        [
            [f'{filename}.{ext}' for ext in extensions]
            for filename in files
        ],
        []
    ))


class Config:
    """
    Основной класс, является частью публичного
    интерфейса библиотеки.

    Пример использования:

    from bestconfig import Config
    config = Config()
    print(config.limit)
    """

    def __new__(cls, *args, exclude_default=False, raise_on_absent=False, exclude: list = None) -> ConfigProvider:
        # Добавить значения по умолчанию
        targets = cls._get_targets(*args, exclude_default=exclude_default, exclude=exclude or set())
        # Передаем файл, из которого был совершен вызов Config()
        # в объекте traceback.extract_stack, последний вызов это данная функция, а перед ним, вызывающая
        resolver = SourceResolver(caller_path=FilesScanner.get_caller_path())
        # Преобразует все цели в один словарь
        config_dict = resolver.resolve_all(targets)
        return ConfigProvider(config_dict)

    """Начало и конец списка источников конфигов, те, что ближе к концу 
    при коллизии перезаписывают более ранние"""
    _targets_begin = generate_targets(applicant_files, supported_extensions)
    _targets_end = ['.env', 'env_file', Source.env]

    @classmethod
    def _get_targets(cls, *args, exclude_default, exclude: list) -> t.List[TargetType]:
        """Дополняет переданные пользователем источники (файлы) стандартными"""
        exclude_set = set(exclude)
        targets = []
        if not exclude_default:
            targets += [target for target in cls._targets_begin if target not in exclude_set]
        targets += args
        if not exclude_default:
            targets += [target for target in cls._targets_end if target not in exclude_set]

        return targets




from abc import ABCMeta, abstractmethod
import typing as t
import ast


class AbstractConverter(metaclass=ABCMeta):
    """
    Наследники данного класса являются преобразователями
    типов данных.
    Например BoolConverter.cast('true') == True
    """

    @abstractmethod
    def cast(self, value: t.Any, safe=True):
        """Принимает переменную любого типа
        Пытается сделать преобразование к определенному типу
        Если safe == False, то при ошибки бросается исключение
        Если safe == True, то при невозможности выполнить
        преобразование, None
        """
        pass


class SimpleConverter(AbstractConverter):
    """Самый простой преобразователь.
    Пытается закастовать к переданному в конструкторе типу
    Пример:
    SimpleConverter(target_type).cast(value) -> target_type(value)
    SimpleConverter(int).cast('123') == 123
    None при неудаче
    """

    def __init__(self, convert_type: type):
        self._target_type = convert_type

    def cast(self, value: t.Any, safe=True):
        if isinstance(value, self._target_type):
            return value

        try:
            return self._target_type(value)
        except (ValueError, TypeError):
            if not safe:
                raise
            return None


class PythonicConverter(AbstractConverter):
    """Конвертирует, исходя из питоновского синтекса.
    Например PythonicConverter("['hi', 234]") = ['hi', 234]
    """

    def __init__(self, empty_as=None):
        """
        :param empty_as: возвращается, если value пустая строка
        """
        self._empty_as = empty_as

    def cast(self, value: t.Any, safe=True):
        if value == '' and self._empty_as is not None:
            return self._empty_as
        if not isinstance(value, str):
            return value
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            if not safe:
                raise
            return None


class UniversalConverter(AbstractConverter):
    """
    Универсальный преобразователь
    1. Если тип значения изначально не строковый, оно возвращается as-is
    2. Производится попытка интерпретировать строку как выражение на python и вернуть его
    3. Возвращается сама строка
    """

    def cast(self, value: t.Any, safe=True):
        if not isinstance(value, str):
            return value
        try:
            return PythonicConverter().cast(value, safe=False)
        except:
            if not safe:
                raise
            return value


class BoolConverter(AbstractConverter):
    """
    'yes' -> True
    'false' -> False
    '0' -> False
    """

    def cast(self, value: t.Any, safe=True) -> bool:
        if isinstance(value, str):
            value = value.lower().strip()
            return self._mapping_dict.get(value, bool(value))

        return bool(value)

    _mapping_dict = {
        'yes': True,
        'no': True,
        'true': True,
        'false': False,
        '1': True,
        '0': False,
        'on': True,
        'off': False
    }



import os
import traceback
import typing as t
from pathlib import Path

from bestconfig.adapters import EnvAdapter, FileAdapter, DictAdapter

from bestconfig.source import Source, TargetType, SourceType


class SourceResolver:
    """
    Получает сырые данные:
    1. Названия файлов
    2. python словари
    3. Source.env, как указание на переменные окружения
    И превращает их в готовые словари
    """

    def __init__(self, caller_path: str):
        self._caller_path = caller_path

    def resolve(self, target: TargetType) -> dict:
        sources = SourceFilter.transform(target, caller_path=self._caller_path)
        aggregator = ConfigAggregator(sources)
        config_dict = aggregator.to_dict()
        return config_dict

    def resolve_all(self, targets: t.List[TargetType]) -> dict:
        data = {}
        for target in targets:
            data.update(self.resolve(target))

        return data


class SourceFilter:
    """Делает preprocessing входных источников,
    превращает их в реальные пути к файлам, удаляет не валид и прочее"""

    @staticmethod
    def _source_from_target(target: TargetType) -> Source:
        """Преобразование пользовательского типа
        в класс Source, отображающий универсальный источник данных"""
        if target == Source.env:
            source = Source(SourceType.ENV)
        elif isinstance(target, dict):
            source = Source(SourceType.DICT)
            source.set('data', target)
        elif isinstance(target, str):
            source = Source(SourceType.FILE)
            source.set('filename', target)
        elif isinstance(target, Path):
            source = Source(SourceType.FILE)
            source.set('filename', str(target))
        else:
            raise ValueError('Unknown source type %s' % type(target))

        return source

    @classmethod
    def transform(cls, target: TargetType, caller_path: str) -> t.List[Source]:
        """Основной метод класса
        Преобразует TargetType -> Source
        То есть названия, алиасы и прочее превращаются в структурированные
        данные, возможно в список, если по паттерну было найдено несколько
        файлов
        """
        clear_sources = []
        scanner = FilesScanner(caller_path=caller_path)
        source = cls._source_from_target(target)

        if SourceType.FILE == source.source_type:
            found_files = scanner.find_all_files(source.get('filename'))
            for filename in found_files:
                new_source = Source(SourceType.FILE)
                new_source.set('filepath', filename)
                clear_sources.append(new_source)
        else:
            clear_sources.append(source)

        return clear_sources


class FilesScanner:
    """
    Задача класса находить файлы конфигурации
    в директориях проекта
    """

    def __init__(self, caller_path: str):
        self._root_path = Path(os.getcwd())
        self._caller_path = caller_path

    @staticmethod
    def get_caller_path():
        """Возвращает файл, в котором была вызвана
        функция, которая вызвала эту"""
        return traceback.extract_stack()[-3][0]

    def find_all_files(self, filename: str) -> t.List[Path]:
        """
        Принимает название или часть названия файла
        и возвращает список подходящих в порядке
        более глубокой вложенности
        """
        paths = []
        curr_dirname = os.path.dirname(self._caller_path)
        curr_dirname = Path(curr_dirname)
        for i in range(self._depth_limit):
            path = self._get_path(curr_dirname, filename)
            if path:
                paths.append(path)
            if curr_dirname == self._root_path:
                break

            curr_dirname = curr_dirname.parent

        passed_file = Path(filename)
        if passed_file.is_file() and passed_file not in paths:
            paths.append(passed_file)

        # Самый последний, наиболее глубокий в файловой структуре
        return list(reversed(paths))

    @classmethod
    def _get_path(cls, dir_path: Path, filename: str) -> t.Optional[Path]:
        """Проверяет в данной директории наличие файла
        по паттерну filename, возвращает None при отсутствии
        и Path() если существует"""
        if not dir_path.exists():
            return None

        filepath = os.path.join(dir_path, filename)
        # Обрабатывает случаи вида /hello/../other
        filepath = Path(os.path.abspath(filepath))
        if not filepath.is_file():
            return None
        return filepath

    _depth_limit = 4


class ConfigSourceAdapter:
    """
    Предоставляет ряд универсальных методов для работы с источником конфига.
    Инкапсулирует реальный доступ к источнику
    Парсинг, модификация, доступ к полям и прочее
    """

    def __init__(self, source: Source):
        self._source = source
        self._adapter = self._adapters_dict[source.source_type]

    def get_dict(self) -> dict:
        return self._adapter.get_dict(self._source)

    _adapters_dict = {
        SourceType.ENV: EnvAdapter,
        SourceType.FILE: FileAdapter,
        SourceType.DICT: DictAdapter
    }


class ConfigAggregator:
    """Превращает сырые словари из файлов и других источников
     в итоговый набор конфигов для пользования"""

    def __init__(self, source: t.List[Source]):
        self._sources = source

    def to_dict(self) -> dict:
        """Возвращает готовый итоговый словарь, содержащий
        все необходимые данные (переменные конфигурации)"""
        return self._combine_sources()

    @classmethod
    def _extract_source(cls, source) -> dict:
        """Возвращает соответствующий источнику словарь данных"""
        adapter = ConfigSourceAdapter(source)
        return adapter.get_dict()

    def _combine_sources(self) -> dict:
        """Возвращает общий для всех источников словарь"""
        data = {}
        for source in self._sources:
            data.update(self._extract_source(source))

        return data



import configparser
import json
import re
from abc import ABCMeta, abstractmethod
from types import ModuleType
from warnings import warn

import yaml

# Название библиотеки, используется для проверок на корректность парсинга .py файлов
LIB_NAME = 'bestconfig'


class AbstractFileParser(metaclass=ABCMeta):
    """
    Родительский класс для парсеров конфигов различных форматов
    """

    """
    Расширение файла, которое обрабатывает данный класс,
    должно переопределяться в наследниках
    """
    extension = ''

    @classmethod
    @abstractmethod
    def read(cls, filepath: str) -> dict:
        """
        :param filepath: полный путь до файла соответствующего типа (extension)
        :return: считанные из файла данные в виде python словаря
        Возбуждает FileNotFoundError при отсутствии файла
        Бросает SyntaxError если файл не соответствует формату
        """
        pass


class YamlParser(AbstractFileParser):
    """Парсит файлы с расширением .yaml"""
    extension = 'yaml'

    @classmethod
    def read(cls, filepath: str) -> dict:
        with open(filepath, 'r') as file:
            try:
                return yaml.load(file, Loader=yaml.Loader)
            except yaml.YAMLError:
                raise SyntaxError


class JsonParser(AbstractFileParser):
    """Парсит файлы с расширением .json"""
    extension = 'json'

    @classmethod
    def read(cls, filepath: str) -> dict:
        with open(filepath, 'r') as file:
            try:
                return json.load(file)
            except json.JSONDecodeError:
                raise SyntaxError


class IniParser(AbstractFileParser):
    """Основана на configparser (https://docs.python.org/3/library/configparser.html)
    Парсит файлы с расширениями .ini или .cfg
    Считывает файл и предоставляет словарь с ключами - секциями
    Пример:
    # config.cfg
    [bitbucket.org]
    User = hg

    [topsecret.server.com]
    Port = 50022
    ForwardX11 = no

    # main.py
    config = Config()
    config.get('topsecret.server.com').Port
    """
    extension = 'ini'

    @classmethod
    def read(cls, filepath: str) -> dict:
        parser = configparser.ConfigParser()
        # Read file with case sensitive keys
        parser.optionxform = str
        parser.read(filepath)
        sections = parser.sections()
        return {
            section_name: dict(parser[section_name])
            for section_name in sections
        }


class EnvParser(AbstractFileParser):
    """Читает .env файлы и подобные ему.
    Файл должен быть в формате VAR_NAME=VAR_VALUE"""

    extension = 'env'

    @classmethod
    def read(cls, filepath: str) -> dict:
        """
        Считывает файл в формате VAR=VALUE в словарь
        Пропускает одинарные и двойные кавычки
        Пропускает комментарии
        """
        template = re.compile(r'''\s*^([^\s#=]+)\s*=\s*(?:[\s"']*)(.*?)(?:[\s"']*)$''')
        result = {}
        with open(filepath, 'r') as ins:
            for line in ins:
                match = template.match(line)
                if match is not None:
                    result[match.group(1)] = match.group(2)
        return result


class PyParser(AbstractFileParser):
    """
    ЭКСПЕРЕМЕНТАЛЬНАЯ ФИЧА
    Файл выполняется и переменные
    из него возвращаются в виде словаря ключ: значение
    Внимание: файл не должен иметь зависимостей, выполняется код
    функцией exec. Если в нем есть import, поведение не определено
    """

    extension = 'env'

    @classmethod
    def read(cls, filepath: str) -> dict:
        with open(filepath, 'r') as file:
            data = file.read()

        if LIB_NAME in data:
            raise AssertionError(
                'Нельзя индексировать .py файла, в котором уже есть Config(), используйте config.update_from_locals()'
            )

        __locals = {}
        data += '\n__locals__ = locals()'
        try:
            # Выполнение кода из файла
            exec(data, __locals, __locals)
        except:
            raise SyntaxError(
                'Ошибка выполнения файла %s, скорее всего в нем есть import-ы, которые приводят к ошибки' % filepath)

        # Оставляем только локальные переменные
        keys = __locals.keys() - {'__builtins__', '__locals__'}

        # Возвращаем все локальные переменные, кроме модулей
        return {
            key: __locals[key]
            for key in keys
            if not isinstance(__locals[key], ModuleType)
        }



import pathlib

from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='bestconfig',
    packages=find_packages(include=['bestconfig']),
    version='1.3.1',
    description="""Setup your project config easily""",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/fivol/bestconfig",
    author='Boris Bondarenko',
    author_email="borisoffficial@gmail.com",
    license='MIT',
    install_requires=[
        'PyYAML>=5.0.0',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
)



from enum import Enum, auto
from pathlib import Path
import typing as t


class SourceType(Enum):
    """Тип переданного в Config() источника """
    ENV = auto()
    DICT = auto()
    FILE = auto()


"""
Тип объекта, передаваемого пользователем для инициализации
словаря конфигурации, это может быть 
1. Название файла (полный пусть, часть пути, относительный путь)
2. Непосредственно словарь
"""
TargetType = t.Union[str, dict, Path]


class Source:
    """
    Класс для хранения промежуточных результатов
    получения итогового словаря из target.
    Соглашение о данных
    filename: str. Путь до файла, относительный или абсолютный
    filepath: Path. Абсолютный путь до файла
    data: dict. Словарь конфигов
    """
    env = '__ENV__'

    def __init__(self, source_type: SourceType):
        """
        :param source_type: тип источника
        """

        # dict в котором хранятся данные, характеризующие источник
        self._data = {}
        self.source_type = source_type

    def set(self, item: str, value):
        self._data[item] = value

    def get(self, item: str):
        return self._data[item]

    def __repr__(self):
        return f'Source({self.source_type})'
